/*
 *  Copyright 2001 Adrian Thurston <adriant@ragel.ca>
 */

/*  This file is part of Ragel.
 *
 *  Ragel is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Ragel is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Ragel; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

#include "rl.h"
#include "ftabcodegen.h"

/* Registers all the section names and inits the code and data pointers. */
FTabCodeGen::FTabCodeGen( char *fsmName, FsmMachine<int> *machine, 
		ParseData *parseData ) : TabCodeGen(fsmName, machine, parseData)
{
	RegisterSection( "FUNC_SWITCH", ftab_FUNC_SWITCH );
	RegisterSection( "STATES", ftab_STATES );
	RegisterSection( "TRANSITIONS", ftab_TRANSITIONS );
}


void ftab_FUNC_SWITCH( FsmCodeGen *codeGen, ostream &out )
{
	int numTransFuncs = codeGen->machine->numTransFuncs;
	int *allTransFuncs = codeGen->machine->allTransFuncs;
	/* Loop over all funcs. */
	int globIndex = 1, funcIndex = 0;
	while ( funcIndex < numTransFuncs ) {
		/* 	We are at the start of a glob, write the case. */
		out.form( "\tcase %d: {\n", globIndex );

		/* Loop over the function list to the start of the next glob. */
		while ( allTransFuncs[funcIndex] != 0 ) {
			out << "{" << codeGen->GetCodeBuiltin( allTransFuncs[funcIndex] ) << "}\n";
			funcIndex++;
		}
		funcIndex++; /* Skip the null terminator. */

		out << "break;}" << endl;

		/* Next glob. */
		globIndex++;
	}
}

void ftab_STATES( FsmCodeGen *codeGen, ostream &out )
{
	int curIndex = 0;
	for ( int i = 0; i < codeGen->machine->numStates; i++ ) {
		FsmMachState<int> *state = codeGen->machine->allStates+i;

		out << "\t{ ";
		if ( state->lowIndex == MAX_INT )
			out << "RAGEL_MAX_INT";
		else
			out << state->lowIndex;
		out << ", ";
		if ( state->highIndex == MIN_INT )
			out << "RAGEL_MIN_INT";
		else
			out << state->highIndex;
		out << ", ";

		if ( state->transIndex )
			out.form( "i+%i, ", curIndex );
		else
			out << "0, ";

		ASSERT( state->errIndex );
		out.form( "%i, ", state->errIndex - codeGen->machine->allTrans );

		out.form("%i, ", (int) state->outFuncs );

		out.form( "%i }", state->isFinState );
		if ( i < codeGen->machine->numStates-1 )
			out << ",\n";

		curIndex += state->numIndex;
	}
}

void ftab_TRANSITIONS( FsmCodeGen *codeGen, ostream &out )
{
	out << '\t';
	FsmMachTrans<int> *trans = codeGen->machine->allTrans;
	for ( int i = 0; i < codeGen->machine->numTrans; i++, trans++ ) {
		if ( trans->toState == 0 )
			out << "{0, ";
		else
			out.form( "{s+%i, ", trans->toState - codeGen->machine->allStates );

		out.form( "%i}", (int) trans->funcs );

		if ( i < codeGen->machine->numTrans-1 ) {
			out << ", ";

			/* Put in a break every 4, but only if not the last. */
			if ( i % 4 == 3 )
				out << "\n\t";
		}
	}
}

/* Registers all the section names and inits the code and data pointers. */
CFTabCodeGen::CFTabCodeGen( char *fsmName, FsmMachine<int> *machine, 
		ParseData *parseData ) : FTabCodeGen(fsmName, machine, parseData)
{
	FsmCodeGen::header = header;
	FsmCodeGen::code = code;
}


char CFTabCodeGen::header[] = 
"
/* Forward dec state for the transition structure. */
struct @FSMNAME@StateStruct;

/* A single transition. */
struct @FSMNAME@TransStruct
{
	struct @FSMNAME@StateStruct *toState;
	int funcs;
};
typedef struct @FSMNAME@TransStruct @FSMNAME@Trans;

/* A single state. */
struct @FSMNAME@StateStruct
{
	int lowIndex;
	int highIndex;
	void *transIndex;
	int errIndex;
	int outFuncs;
	int isFinState;
};
typedef struct @FSMNAME@StateStruct @FSMNAME@State;

/* Only non-static data: current state. */
struct @FSMNAME@Struct
{
	@FSMNAME@State *curState;
	int accept;
	@STRUCT_DATA@
};
typedef struct @FSMNAME@Struct @FSMNAME@;

/* Init the fsm. */
void @FSMNAME@Init( @FSMNAME@ *fsm );

/* Execute some chunk of data. */
void @FSMNAME@Execute( @FSMNAME@ *fsm, char *data, int len );

/* Indicate to the fsm tha there is no more data. */
void @FSMNAME@Finish( @FSMNAME@ *fsm );

/* Did the machine accept? */
int @FSMNAME@Accept( @FSMNAME@ *fsm );

";

char CFTabCodeGen::code[] = "

#define RAGEL_MAX_UINT (~((unsigned int)0))
#define RAGEL_MAX_INT ((int)(RAGEL_MAX_UINT>>1))
#define RAGEL_MIN_INT ((int)((RAGEL_MAX_UINT>>1)+1))

static @FSMNAME@State @FSMNAME@_s[];
static @INDEX_TYPE@ @FSMNAME@_i[];
static @FSMNAME@Trans @FSMNAME@_t[];

#define s @FSMNAME@_s
#define i @FSMNAME@_i
#define t @FSMNAME@_t

/* The aray of states. */
static @FSMNAME@State @FSMNAME@_s[] = {
@STATES@
};

/* The array of indicies into the transition array. */
#if @ANY_INDICIES@
static @INDEX_TYPE@ @FSMNAME@_i[] = {
@INDICIES@
};
#endif

/* The array of trainsitions. */
static @FSMNAME@Trans @FSMNAME@_t[] = {
@TRANSITIONS@
};

/* The start state. */
static @FSMNAME@State *@FSMNAME@_startState = s+@START_STATE_OFFSET@;
static @FSMNAME@Trans *@FSMNAME@_errIndex = t;

/* Global low and high indicies. */
static int @FSMNAME@_gblLowIndex = @GBL_LOW_INDEX@;
static int @FSMNAME@_gblHighIndex = @GBL_HIGH_INDEX@;

#undef f
#undef s
#undef i
#undef t


/***************************************************************************
 * @FSMNAME@ExecFuncs
 *
 * Execute functions pointed to by funcs until the null function is found. 
 */
inline static void @FSMNAME@ExecFuncs( @FSMNAME@ *fsm, int funcs, int c )
{
	switch ( funcs ) {
@FUNC_SWITCH@
	}
}

/****************************************
 * @FSMNAME@Init
 */
void @FSMNAME@Init( @FSMNAME@ *fsm )
{
	fsm->curState = @FSMNAME@_startState;
	fsm->accept = 0;
	@INIT_CODE@
}

/****************************************
 * @FSMNAME@Accept
 *
 * Did the fsm accept? 
 */
int @FSMNAME@Accept( @FSMNAME@ *fsm )
{
	return fsm->accept;
}


/**********************************************************************
 * @FSMNAME@Execute
 *
 * Execute the fsm on some chunk of data. 
 */
void @FSMNAME@Execute( @FSMNAME@ *fsm, char *data, int len )
{
	@FSMNAME@State *cs = fsm->curState;
	for ( ; len > 0; data++, len-- ) {
		int c = *(unsigned char*)data;
		@FSMNAME@Trans *trans;

		if ( cs == 0 )
			goto finished;

		if ( @FSMNAME@_gblLowIndex <= c && c < @FSMNAME@_gblHighIndex ) {
			/* If the character is within the index bounds then get the
			 * transition for it. If it is out of the transition bounds
			 * we will use the error transition. */
			if ( cs->lowIndex <= c && c < cs->highIndex ) {
				/* Use the index to look into the transition array. */
				trans = @FSMNAME@_t + 
					((@INDEX_TYPE@*)cs->transIndex)[c - cs->lowIndex];
			}
			else {
				/* Use the error index as the char is out of range. */
				trans = @FSMNAME@_t + cs->errIndex;
			}
		}
		else {
			trans = @FSMNAME@_errIndex;
		}

		@PRE_FUNC_LIST@

		/* If there are functions for this transition then execute them. */
		if ( trans->funcs )
			@FSMNAME@ExecFuncs( fsm, trans->funcs, c );

		@POST_FUNC_LIST@

		/* Move to the new state. */
		cs = trans->toState;
	}
finished:
	fsm->curState = cs;
}

/**********************************************************************
 * @FSMNAME@Finish
 *
 * Indicate to the fsm that the input is done. Does cleanup tasks.
 */
void @FSMNAME@Finish( @FSMNAME@ *fsm )
{
	@FSMNAME@State *cs = fsm->curState;
	if ( cs != 0 && cs->isFinState ) {
		/* If finishing in a final state then execute the
		 * out functions for it. (if any). */
		if ( cs->outFuncs != 0 )
			@FSMNAME@ExecFuncs( fsm, cs->outFuncs, -1 );
		fsm->accept = 1;
	}
	else {
		/* If we are not in a final state then this
		 * is an error. Move to the errror state. */
		fsm->curState = 0;
	}
}
";


/* Registers all the section names and inits the code and data pointers. */
CCFTabCodeGen::CCFTabCodeGen( char *fsmName, FsmMachine<int> *machine, 
		ParseData *parseData ) : FTabCodeGen(fsmName, machine, parseData)
{
	FsmCodeGen::header = header;
	FsmCodeGen::code = code;
}

char CCFTabCodeGen::header[] = 
"

class @FSMNAME@
{
public:
	/* Function and index type. */
	typedef int Func;

	/* Forward dec state for the transition structure. */
	struct State;

	/* A single transition. */
	struct Trans
	{
		State *toState;
		int funcs;
	};

	/* A single state. */
	struct State
	{
		int lowIndex;
		int highIndex;
		void *transIndex;
		int errIndex;
		int outFuncs;
		int isFinState;
	};

	/* Constructor. */
	@FSMNAME@();

	void Init( );

	/* Execute some chunk of data. */
	void Execute( char *data, int len );

	/* Indicate to the fsm tha there is no more data. */
	void Finish( );

	/* Did the machine accept? */
	int Accept( );

	State *curState;
	int accept;

	inline void ExecFuncs( int funcs, int c );

	@STRUCT_DATA@
};


";

char CCFTabCodeGen::code[] = "

#define RAGEL_MAX_UINT (~((unsigned int)0))
#define RAGEL_MAX_INT ((int)(RAGEL_MAX_UINT>>1))
#define RAGEL_MIN_INT ((int)((RAGEL_MAX_UINT>>1)+1))

static @FSMNAME@::State @FSMNAME@_s[];
static @INDEX_TYPE@ @FSMNAME@_i[];
static @FSMNAME@::Trans @FSMNAME@_t[];

#define s @FSMNAME@_s
#define i @FSMNAME@_i
#define t @FSMNAME@_t

/* The aray of states. */
static @FSMNAME@::State @FSMNAME@_s[] = {
@STATES@
};

/* The array of indicies into the transition array. */
#if @ANY_INDICIES@
static @INDEX_TYPE@ @FSMNAME@_i[] = {
@INDICIES@
};
#endif

/* The array of trainsitions. */
static @FSMNAME@::Trans @FSMNAME@_t[] = {
@TRANSITIONS@
};

/* The start state. */
static @FSMNAME@::State *@FSMNAME@_startState = s+@START_STATE_OFFSET@;
static @FSMNAME@::Trans *@FSMNAME@_errIndex = t;


/* Global low and high indicies. */
int @FSMNAME@_gblLowIndex = @GBL_LOW_INDEX@;
int @FSMNAME@_gblHighIndex = @GBL_HIGH_INDEX@;



/***************************************************************************
 * Execute functions pointed to by funcs until the null function is found. 
 */
inline void @FSMNAME@::ExecFuncs( int funcs, int c )
{
	switch ( funcs ) {
@FUNC_SWITCH@
	}
}


/****************************************
 * Constructor
 */
@FSMNAME@::@FSMNAME@( )
{
	Init();
}

/****************************************
 * @FSMNAME@Init
 */
void @FSMNAME@::Init( )
{
	curState = @FSMNAME@_startState;
	accept = 0;
	@INIT_CODE@
}

/****************************************
 * Did the fsm accept? 
 */
int @FSMNAME@::Accept( )
{
	return accept;
}



/**********************************************************************
 * Execute the fsm on some chunk of data. 
 */
void @FSMNAME@::Execute( char *data, int len )
{
	State *cs = curState;
	for ( ; len > 0; data++, len-- ) {
		int c = *(unsigned char*)data;
		Trans *trans;

		if ( cs == 0 )
			goto finished;

		if ( @FSMNAME@_gblLowIndex <= c && c < @FSMNAME@_gblHighIndex ) {
			/* If the character is within the index bounds then get the
			 * transition for it. If it is out of the transition bounds
			 * we will use the error transition. */
			if ( cs->lowIndex <= c && c < cs->highIndex ) {
				/* Use the index to look into the transition array. */
				trans = @FSMNAME@_t + 
					((@INDEX_TYPE@*)cs->transIndex)[c - cs->lowIndex];
			}
			else {
				/* Use the error index as the char is out of range. */
				trans = @FSMNAME@_t + cs->errIndex;
			}
		}
		else {
			trans = @FSMNAME@_errIndex;
		}

		@PRE_FUNC_LIST@

		/* If there are functions for this transition then execute them. */
		if ( trans->funcs != 0 )
			ExecFuncs( trans->funcs, c );

		@POST_FUNC_LIST@

		/* Move to the new state. */
		cs = trans->toState;
	}
finished:
	curState = cs;
}


/**********************************************************************
 * Indicate to the fsm that the input is done. Does cleanup tasks.
 */
void @FSMNAME@::Finish( )
{
	State *cs = curState;
	if ( cs != 0 && cs->isFinState ) {
		/* If finishing in a final state then execute the
		 * out functions for it. (if any). */
		if ( cs->outFuncs != 0 )
			ExecFuncs( cs->outFuncs, -1 );
		accept = 1;
	}
	else {
		/* If we are not in a final state then this
		 * is an error. Move to the errror state. */
		curState = 0;
	}
}

";

